//=============================================================================================
// A BWT class using the MSufSort suffix sorting algorithm
//
// Author: M.A. Maniscalco
// Date: 7/30/04
// email: michael@www.michael-maniscalco.com
//
// This code is free for non commercial use only.
//
//=============================================================================================




#include "BWT.h"
#include <conio.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

#define VERBOSE						




BWT::BWT()
{
	m_suffixSorter = new MSufSort;
}




BWT::~BWT()
{
	delete m_suffixSorter;
}





unsigned int BWT::Forward(unsigned char * data, unsigned int length)
{
	//------------------------------------------------------------------------------
	// 1.	Do the suffix sort.
	//------------------------------------------------------------------------------
	#ifdef VERBOSE
		printf("Doing MSufSort ...%c", 13);
	#endif

	int start = clock();
	unsigned int key = m_suffixSorter->Sort(data, length);
	
	#ifdef CHECK_SORT
		if (!m_suffixSorter->VerifySort())
			printf("\n*** Sort error detected ***\n");
		else
			printf("\nSort Verified\n");
	#endif

	#ifdef VERBOSE
		printf("Doing BWT ...       %c", 13);
	#endif

	//-----------------------------------------------------------------------------
	// 2.	Calculate the BWT from the suffix sort.
	//		This could also be done within the OnSortedSuffix() method during the
	//		suffix sort.  But is done after so that the MSufSort algorithm itself can
	//		be timed without the BWT portion.
	//-----------------------------------------------------------------------------

	unsigned char * bwtBuffer = new unsigned char[length];
	for (unsigned int i = 1; i <= length; i++)
	{
		unsigned int sortedPosition = m_suffixSorter->ISA(i);
		if (sortedPosition >= key)
			sortedPosition--;
		bwtBuffer[sortedPosition] = data[i - 1];
	}
	memcpy(data, bwtBuffer, length);
	int finish = clock();

	#ifdef VERBOSE
		double elapsedTime = (double)(finish - start) / 1000;
		double msufsortTime = (double)m_suffixSorter->GetElapsedSortTime() / 1000;
		printf("    Elapsed time: %.3f seconds\n", elapsedTime);
		printf("    MSufSort time: %.3f seconds\n", msufsortTime);
		printf("    BWT time: %.3f seconds\n", elapsedTime - msufsortTime);
		printf("\n");
	#endif
	
	delete [] bwtBuffer;
	return key;
}





	
void BWT::Reverse(unsigned char * data, unsigned int length, unsigned int forwardKey)
{
	// reverses the blocksort transform
	#ifdef VERBOSE
		printf("Reversing the BWT ...%c", 13);
	#endif

	int start = clock();

	unsigned int * index = new unsigned int[length + 1];
	unsigned int symbolRange[257];
	unsigned char * ptr;

	memset(symbolRange, 0, sizeof(unsigned int) * 257);
	ptr = data;

	symbolRange[0] = 1;		// our -1 symbol
	for (unsigned int i = 0; i < length; i++, ptr++)
		symbolRange[(*ptr) + 1]++;

	unsigned int n = 0;
	for (i = 0; i < 257; i++)
	{
		unsigned int temp = symbolRange[i];
		symbolRange[i] = n;
		n += temp;
	}

	n = 0;
	index[0] = forwardKey;
	for (i = 0; i < length; i++, n++)
	{
		if (i == forwardKey)
			n++;
		index[symbolRange[data[i] + 1]++] = n;
	}


	n = forwardKey;
	char * reversedBuffer = new char [length];
	for (i = 0; i < length; i++)
	{
		n = index[n];
		if (n >= forwardKey)
			reversedBuffer[i] = data[n - 1];
		else
			reversedBuffer[i] = data[n];
	}

	memcpy(data, reversedBuffer, length);
	#ifdef USE_ALT_SORT_ORDER
	MSufSort::ReverseAltSortOrder(data, length);
	#endif

	int finish = clock();

	#ifdef VERBOSE
		printf("Reverse BWT finished.  Elapsed time: %.2f seconds\n", (double)(finish - start) / 1000);
	#endif

	delete [] index;
	delete [] reversedBuffer;
}
